/***************************************************************************
 *   Copyright (C) 2006 by Jonas Gehring   *
 *   saubue@boolsoft.org   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>


#define CREDITS     "\nASM to C Hex - Converter\nCopyLeft 2006 by Jonas Gehring\n\n"

const char *forms[4] = {"unsigned char", "unsigned short", "", "unsigned long"};

FILE *inf;
FILE *outf;

short firstentry, noarr;

// Output: How to use the Program
void PrintUsage(void)
{
    printf(CREDITS);
    printf("Usage: asmtochex [flags] <type> <infile> <outfile>\n\n"             \
            "       flags: -quiet    -  Don't ouput standard messages\n"         \
            "              -noarr    -  Don't use standard array format\n\n"     \
            "       types: -uchar    -  Convert to unsigned char array\n"        \
            "              -ushort   -  Convert to unsigned short array\n"       \
            "              -ulong    -  Convert to unsigned long array\n\n"      \
            "              infile    -  Name of input file\n"                    \
            "              outfile   -  Name of output file\n\n"                 \
            "              Converts 68k ASM data to C Hex arrays\n"              \
            "              of given type.\n\n" );
}


// Replace " dc.b "-beginnings and ";"-Comments
unsigned char *ReplaceBeginn(unsigned char *buffer)
{   
    if (!strncmp(buffer, " dc.b ", 6)) {
        buffer += 6;
    }    
    else if (!strncmp(buffer, " ;", 2)) { 
        fputs("// ", outf);
        buffer += 2;
    }    
    else if (!strncmp(buffer, ";", 1)) {
        fputs("// ", outf);
        buffer++;
    }
    else if (strncmp(buffer, ",", 1) || strncmp(buffer, "$", 1)) {
        if (firstentry == -2) firstentry = -1;
        buffer = NULL;
    }    
    
    /*if (!strncmp(bptr, "tileData:", 9) || !strncmp(bptr, "mapData:", 8)){
    buffer = NULL;   
}*/        
                       
    return buffer;
}


// Convert to given format
unsigned long long ConvertData(const char *infile, short flag)
{
    unsigned char buffer[65536];
    unsigned char *bptr;
    unsigned long i, line = -1;
    unsigned long long numel = 0;
    short n = 0;
    
    firstentry = -2;
    
    // Count the number of entries
    while (fgets(buffer, 65535, inf)) {
        for (i = 0; i < strlen(buffer); i++) {                                                                      
            if (!strncmp(&buffer[i], "$", 1)) {
                numel++;
            } 
        }
    }
    
    numel = numel/flag;
    
    // Close the file and open it again (so we are at
    // the beginning of the file again)
    fclose(inf);
    if (!(inf = fopen(infile, "r")))
        return 0;    
    
    // Converting loop (read every line)
    while (fgets(buffer, 65535, inf)) {
        bptr = ReplaceBeginn(buffer);
                   
          //if (!strncmp(bptr, "tileData:", 9) || !strncmp(bptr, "mapData:", 8)) {
        if (firstentry == -1) {
            if (!noarr)          fprintf(outf, "%s data[%d] = {\n", forms[flag-1], numel);
            firstentry = 0;
            continue;         
        }
        
          // If there's nothing to convert
        if (bptr == NULL)
            continue;  
        
        for (i = 0; i < strlen(bptr); i++) {                                                                      
              // Remove "$"-Prefixes and replace them with "0x"
            if (!strncmp(&bptr[i], "$", 1)) {
                n++;
                          
                if (n/flag == 1 || !i) {                  
                    n = 0; 
                      // Write "0x" at the beginning of an entry
                    if (!firstentry) {  fputs(" 0x", outf); firstentry++; }
                    else             {  fputs(",0x", outf); }
                }                              
          
            } else if (strncmp(&bptr[i], ",", 1)) {
                  // Write data to file
                fputc(bptr[i], outf);
            } 
        }
    }
    
    // Put close brackets at the end of the C Hex array
    if (!noarr)   fputs("};", outf);
    
    return numel;
}


// Program entry point
int main(int argc, char *argv[])
{
    char *infile = NULL, *outfile = NULL;
    unsigned long long n;
    short cflag, quiet;
    clock_t prgstart;
    
    prgstart = clock();
    
    quiet = 0;
    noarr = 0;
        
    // Check for arguments
    if (argc < 3) {
        PrintUsage();
        return 0;
    }
    
    // Parse arguments
    for (n = 1; n < argc; n++) {
        if (!strcmp(argv[n], "-uchar"))       cflag       = 1;
        else if (!strcmp(argv[n], "-ushort")) cflag       = 2;
        else if (!strcmp(argv[n], "-ulong"))  cflag       = 4;
        else if (!strcmp(argv[n], "-quiet"))  quiet       = 1;
        else if (!strcmp(argv[n], "-noarr"))  noarr       = 1;
        else if (!infile)                     infile      = argv[n];
        else if (!outfile)                    outfile     = argv[n];
        else {
            PrintUsage();
            return 0;
        }
    }
    if (!infile || !outfile) {
        PrintUsage();
        return 0;
    }
    
    // Try to open input and output file (else return)
    if (!(inf = fopen(infile, "r"))) {
        printf(CREDITS);
        printf("ERROR: unable to open inputfile %s\n", infile);
        return 0;
    }
    if (!(outf = fopen(outfile,"wb"))) {
        printf(CREDITS);
        printf("ERROR: unable to open outputfile %s\n", outfile);
        fclose(inf);
        return 0;
    }
    
    // Now convert the file
    n = ConvertData(infile, cflag);


    // Finally, print the informations (if -quiet is not given)
    printf(CREDITS);
    
    if (!quiet) {
        printf("Data converted in %.3f seconds\nArray type: %s\nNumber of entries: %li\n\n", ((float)(clock() - prgstart) / CLOCKS_PER_SEC), forms[--cflag], n);
    } else {
        printf("Data successfully converted.\n\n");
    }
                 
    return 0;
}
